package application

import com.googlecode.lanterna.terminal.DefaultTerminalFactory
import com.googlecode.lanterna.terminal.IOSafeTerminal
import com.googlecode.lanterna.terminal.IOSafeTerminalAdapter.createRuntimeExceptionConvertingAdapter
import com.googlecode.lanterna.terminal.Terminal
import interpreter.LispInterpreter
import interpreter.LispInterpreterBuilder
import stream.UncheckedIOException
import terminal.LispTerminal
import terminal.LispTerminal.END_OF_SEGMENT
import terminal.TerminalConfiguration
import java.io.IOException
import java.io.PipedInputStream
import java.io.PipedOutputStream
import java.io.PrintStream
import java.text.MessageFormat.format

class LispMain {

    private var builder: LispInterpreterBuilder? = null
    private var inputReader: PipedInputStream? = null
    private var output: PrintStream? = null
    private var lispTerminal: LispTerminal? = null
    private var configuration: TerminalConfiguration? = null

    private val version = LispMain::class.java.`package`.implementationVersion

    constructor() {
        this.builder = LispInterpreterBuilder
    }

    constructor(configuration: TerminalConfiguration) {
        this.builder = LispInterpreterBuilder
        this.configuration = configuration
    }

    fun runInteractive() {
        initializeTerminal()
        printGreeting()
        lispTerminal!!.start()
        buildInteractiveInterpreter().interpret()
        shutdownTerminal()
    }

    private fun initializeTerminal() {
        createTerminalConfiguration()
        inputReader = configuration!!.inputReader
        output = PrintStream(configuration!!.outputWriter)
        lispTerminal = LispTerminal(configuration!!)
    }

    private fun createTerminalConfiguration() {
        if (configuration == null) {
            configuration = TerminalConfiguration()
            configuration!!.setInputPair(PipedOutputStream(), PipedInputStream())
            configuration!!.setOutputPair(PipedOutputStream(), PipedInputStream())
            configuration!!.terminal = createIOSafeTerminal()
        }
    }

    private fun createIOSafeTerminal(): IOSafeTerminal {
        return createRuntimeExceptionConvertingAdapter(createDefaultTerminal())
    }

    private fun createDefaultTerminal(): Terminal {
        try {
            return DefaultTerminalFactory().createTerminal()
        } catch (e: IOException) {
            throw UncheckedIOException(e)
        }
    }

    private fun printGreeting() {
        output!!.println(format(GREETING, (version ?: "null")))
        output!!.println()
    }

    private fun buildInteractiveInterpreter(): LispInterpreter {
        builder!!.setInput(inputReader!!, "terminal")
        builder!!.setOutput(output!!)
        builder!!.setErrorOutput(output!!)
        builder!!.setTerminationFunction { this.shutdownTerminal() }
        builder!!.setErrorTerminationFunction { this.shutdownTerminal() }
        builder!!.setLanguageFileNames(*LANGUAGE_FILE_NAMES)
        builder!!.setPromptDecorator { s -> s + END_OF_SEGMENT }
        builder!!.setValueOutputDecorator(makeColorDecorator(ANSI_GREEN))
        builder!!.setWarningOutputDecorator(makeColorDecorator(ANSI_YELLOW))
        builder!!.setErrorOutputDecorator(makeColorDecorator(ANSI_RED))
        builder!!.setCriticalOutputDecorator(makeColorDecorator(ANSI_PURPLE))

        return builder!!.build()
    }

    private fun shutdownTerminal() {
        output!!.print(END_OF_SEGMENT)
        lispTerminal!!.stop()
        output!!.close()
    }

    private fun makeColorDecorator(color: String): (String) -> String {
        return { color + it + ANSI_RESET }
    }

    fun runWithFile(fileName: String) {
        buildFileInterpreter(fileName).interpret()
    }

    private fun buildFileInterpreter(fileName: String): LispInterpreter {
        builder!!.useFile(fileName)
        builder!!.setOutput(System.out)
        builder!!.setErrorOutput(System.err)
        builder!!.setTerminationFunction { System.exit(0) }
        builder!!.setErrorTerminationFunction { System.exit(1) }
        builder!!.setLanguageFileNames(*LANGUAGE_FILE_NAMES)
        builder!!.setValueOutputDecorator(makeColorDecorator(ANSI_GREEN))
        builder!!.setWarningOutputDecorator(makeColorDecorator(ANSI_YELLOW))
        builder!!.setErrorOutputDecorator(makeColorDecorator(ANSI_RED))
        builder!!.setCriticalOutputDecorator(makeColorDecorator(ANSI_PURPLE))

        return builder!!.build()
    }

    companion object {

        val GREETING = "Transcendental Lisp - Version {0}"

        val ANSI_RESET = "\u001B[0m"
        val ANSI_RED = "\u001B[31m"
        val ANSI_GREEN = "\u001B[32m"
        val ANSI_YELLOW = "\u001B[33m"
        val ANSI_PURPLE = "\u001B[35m"

        val LANGUAGE_FILE_NAMES = arrayOf("functions.lisp", "dlambda.lisp")

        @JvmStatic
        fun main(arguments: Array<String>) {
            val lispMain = LispMain()

            if (arguments.size == 0)
                lispMain.runInteractive()
            else
                lispMain.runWithFile(arguments[0])
        }
    }
}
