package interpreter

import environment.RuntimeEnvironment
import interpreter.InteractiveLispInterpreter.PROMPT
import org.junit.After
import org.junit.Assert.assertEquals
import org.junit.Assert.assertFalse
import org.junit.Assert.assertTrue
import org.junit.Before
import org.junit.Test
import testutil.TestUtilities.createInputStreamFromString
import java.io.ByteArrayOutputStream
import java.io.PrintStream
import java.text.MessageFormat.format
import java.util.HashSet

class LispInterpreterTest {

    private var indicatorSet: MutableSet<String>? = null
    private var outputStream: ByteArrayOutputStream? = null
    private var errorOutputStream: ByteArrayOutputStream? = null
    private val environment: RuntimeEnvironment
    private val builder: LispInterpreterBuilder

    init {
        this.environment = RuntimeEnvironment
        this.builder = LispInterpreterBuilder
    }

    private fun setCommonFeatures() {
        builder.setOutput(PrintStream(outputStream!!))
        builder.setErrorOutput(PrintStream(errorOutputStream!!))
        builder.setTerminationFunction { }
        builder.setErrorTerminationFunction { indicatorSet!!.add(TERMINATED) }
    }

    private fun assertTerminated() {
        assertTrue(indicatorSet!!.contains(TERMINATED))
    }

    private fun assertErrorMessageWritten() {
        assertTrue(errorOutputStream!!.toByteArray().size > 0)
    }

    @Before
    fun setUp() {
        indicatorSet = HashSet()
        outputStream = ByteArrayOutputStream()
        errorOutputStream = ByteArrayOutputStream()
        environment.reset()
        builder.reset()
    }

    @After
    fun tearDown() {
        environment.reset()
        builder.reset()
    }

    @Test
    fun buildInteractiveInterpreter() {
        setCommonFeatures()
        builder.setInput(System.`in`, "stdin")
        val interpreter = builder.build()

        assertTrue(interpreter is InteractiveLispInterpreter)
    }

    @Test
    fun buildNonInteractiveInterpreter() {
        setCommonFeatures()
        builder.setInput(System.`in`, "stdin")
        builder.setNotInteractive()
        val interpreter = builder.build()

        assertFalse(interpreter is InteractiveLispInterpreter)
        assertFalse(interpreter is FileLispInterpreter)
    }

    @Test
    fun buildFileBasedInterpreter() {
        setCommonFeatures()
        builder.useFile(FILE)
        val interpreter = builder.build()

        assertTrue(interpreter is FileLispInterpreter)
    }

    @Test
    fun attemptToBuildInterpreterOnBadFile() {
        setCommonFeatures()
        builder.useFile("does-not-exist.lisp")
        builder.build()

        assertErrorMessageWritten()
        assertTerminated()
    }

    @Test
    fun makeSureDecoratorsAreInitializedWithDefaults() {
        builder.build()

        assertEquals("", environment.decoratePrompt(""))
        assertEquals("", environment.decorateValueOutput(""))
        assertEquals("", environment.decorateWarningOutput(""))
        assertEquals("", environment.decorateErrorOutput(""))
        assertEquals("", environment.decorateCriticalOutput(""))
    }

    @Test
    fun makeSureDecoratorsAreSetCorrectly() {
        builder.setPromptDecorator { s -> "#$s#" }
        builder.setValueOutputDecorator { s -> "@$s@" }
        builder.setWarningOutputDecorator { s -> "%$s%" }
        builder.setErrorOutputDecorator { s -> "*$s*" }
        builder.setCriticalOutputDecorator { s -> "$$s$" }
        builder.build()

        assertEquals("#x#", environment.decoratePrompt("x"))
        assertEquals("@x@", environment.decorateValueOutput("x"))
        assertEquals("%x%", environment.decorateWarningOutput("x"))
        assertEquals("*x*", environment.decorateErrorOutput("x"))
        assertEquals("\$x$", environment.decorateCriticalOutput("x"))
    }

    @Test
    fun fileBasedInterpreterWorks_PrintsLastValueOnly() {
        setCommonFeatures()
        builder.useFile(FILE)
        builder.build().interpret()

        assertEquals("PICKLE\n\n", outputStream!!.toString())
        assertEquals("", errorOutputStream!!.toString())
    }

    @Test
    fun interactiveInterpreterWorks() {
        setCommonFeatures()
        builder.setInput(createInputStreamFromString("'pickle"), "input")
        builder.build().interpret()

        assertEquals(format("{0}\n{1}\n{0}\n", PROMPT, "PICKLE"), outputStream!!.toString())
        assertEquals("", errorOutputStream!!.toString())
    }

    @Test
    fun interpreterHandlesError() {
        setCommonFeatures()
        builder.setNotInteractive()
        builder.setInput(createInputStreamFromString("pickle"), "input")
        builder.build().interpret()

        assertEquals("\n", outputStream!!.toString())
        assertEquals("[error] symbol PICKLE has no value\n", errorOutputStream!!.toString())
    }

    companion object {

        private val TERMINATED = "terminated"
        private val FILE = LispInterpreterTest::class.java.getResource("file.lisp").file
    }
}
